/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#ifndef BITBUFFER_H
#define BITBUFFER_H

#include <bit/enums.h>
#include <bit/data.h>
#include <cstring>

/**
 * @defgroup buffers "Data Buffers"
 * Buffers provide an interface for storage of data that is not necessarily octet aligned.
 */

namespace bit
  {

  /**
   * This class manages a packed bit (or octet) buffer. Although this class
   * manages bit oriented buffers, it attempts to optimize for octet (byte)
   * based buffers.
   *
   * The methods pack, unpack and clear provide a means of extracting
   * bit segments from the buffer.
   *
   * @todo create an overloaded = operator that copies memory if the buffer is the owner
   *
   * @ingroup buffers
   * @author Rick L. Vinyard, Jr.
   */
  class Buffer: public Data
    {
    protected:

      /**
       * Construct a buffer of a given size that is allowed to dynamically grown as
       * pack/unpack requests are made.
       */
      Buffer(size_t initial_size=0, size_t sizemax=0);

      /**
       * Construct a buffer that is a copy of external data.
       */
      Buffer(void* data, size_t size, size_t sizemax=0, DataMode mode=COPY);

    public:

      typedef BitPointer<Buffer> pointer;

      static pointer create(size_t initial_size=0, size_t sizemax=0);

      static pointer create(void* data, size_t size, size_t sizemax=0, DataMode mode=COPY);

      virtual ~Buffer();

      /**
       * Transfer data from the buffer area into the low order bits of a memory location.
       * @param mem The memory location to transfer buffer data into
       * @param mem_octets The size of the memory location in octets
       * @param buf_offset_bits The offset from the beginning of the buffer in bits to begin extraction from
       * @param bits The size of the data to extract in bits
       * @return true if transfer accomplished, false otherwise
       *
       * Bit Buffer
       * ==================================================
       * |000000000011111111112222222222333333333344444444|
       * |012345678901234567890123456789012345678901234567|
       * ==================================================
       *                ^ buf_offset_bits = 14
       *                ^ extract_bits = 20 ^
       *
       * Memory Buffer
       * ==================================================
       * |000000000011111111112222222222333333333344444444|
       * |012345678901234567890123456789012345678901234567|
       * ==================================================
       * ^ mem = start of memory address
       * ^ mem_octets = 6                                ^
       *                             ^ Resulting data    ^ when mem_loc = END
       * ^ Resulting data    ^                             when mem_loc = START
       */
      bool unpack(void* mem, size_t mem_octets, Location mem_loc, size_t buf_offset_bits, size_t extract_bits);

      Data unpack( size_t buf_offset_bits, size_t extract_bits, Location mem_loc=END );

      template <typename T>
      bool unpack(T& val, Location mem_loc, size_t buf_offset_bits, size_t extract_bits)
      {
        return unpack(&val, sizeof(T), mem_loc, buf_offset_bits, extract_bits);
      }

      /**
       * Transfer data from the low order bits of a memory location into the low order bits of a segment of the buffer area.
       * @param mem The memory location to transfer from
       * @param mem_octets The size of the memory location in octets
       * @param buf_offset The offset from the start of the data buffer in bits
       * @param buf_tgtsize The size of the target area in the buffer in bits
       * @param n The size of the data to extract in bits
       * @return true if transfer accomplished, false otherwise.
       * If @param offset + @param destsize are greater than the size of the buffer and the buffer
       * cannot be expanded to accommodate the request zero is returned.
       *
       * Memory Buffer
       * ==================================================
       * |000000000011111111112222222222333333333344444444|
       * |012345678901234567890123456789012345678901234567|
       * ==================================================
       * ^ mem = start of memory address
       * ^ mem_octets = 6                                ^
       *                                        ^ n = 10 ^ when mem_loc = END
       *                                        ^ cpy out^ when mem_loc = END
       * ^ n = 10 ^                                        when mem_loc = START
       * ^ cpy out^                                        when mem_loc = START
       *
       * Bit Buffer
       * ==================================================
       * |000000000011111111112222222222333333333344444444|
       * |012345678901234567890123456789012345678901234567|
       * ==================================================
       *                ^ buf_offset_bits = 14
       *                ^ buf_tgtsize  = 20 ^
       *                          ^ n = 10  ^
       *                          ^ cpy in  ^
       */
      bool pack(const void* mem, size_t mem_octets, Location mem_loc, size_t buf_offset, size_t buf_tgtsize, size_t n);

      bool pack( Data d, size_t buf_offset, size_t buf_tgtsize, size_t n, Location mem_loc=END );

      template <typename T>
      bool pack(const T& val, Location mem_loc, size_t offset, size_t destsize, size_t n)
      {
        bool b;
        b = pack(&val, sizeof(T), mem_loc, offset, destsize, n);
        return b;
      }

      bool clear_bits(size_t offset, size_t bits)
      {
        return clear_bits(offset, bits, false);
      }

      /** Gets the maximum size this buffer is allowed to grow */
      size_t size_max();

      /**
       * Sets the maximum size that this buffer is allowed to grow.
       * Note: if sizemax is set to a value less than the current
       * buffer size the buffer will be truncated immediately.
       */
      void set_size_max(size_t sizemax);

      virtual void set_data( const void* newdata, size_t newsize, DataMode mode=COPY ) throw (std::bad_alloc);

      virtual bool resize( size_t s ) throw (std::bad_alloc);

      /** Signals a change to the data within the buffer. */
      sigc::signal<void> signal_changed()
      {
        return m_signal_changed;
      }

    protected:
      size_t m_size_max;

      /**
       * Utility method to clear bits and suppress calling the on_changed* methods
       */
      bool clear_bits(size_t offset, size_t bits, bool suppress);

      virtual void on_data_changed();

    private:
      sigc::signal<void> m_signal_changed;

      bool m_signal_changed_need_reemit;
      bool m_signal_changed_emitting;


    };

}

#endif
