/***************************************************************************
*   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
*   rvinyard@cs.nmsu.edu                                                  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU Lesser General Public License as        *
*   published by the Free Software Foundation version 2.1.                *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU Lesser General Public      *
*   License along with this library; if not, write to the                 *
*   Free Software Foundation, Inc.,                                       *
*   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
***************************************************************************/
#ifndef BITFIELDBUFFER_H
#define BITFIELDBUFFER_H

#include <bit/fieldbase.h>
#include <bit/data.h>

#include <sstream>

namespace bit
  {

  class RecordBuffer;


  /**
   * Combines the concept of a field with a means of extracting field
   * data from a buffer.
   *
   * @ingroup buffers
   * @author Rick L. Vinyard, Jr. <rvinyard@cs.nmsu.edu>
   */
  class FieldBuffer
    {
    public:

      FieldBuffer(RecordBuffer& b, FieldBase::pointer f);

      virtual ~FieldBuffer();

      const std::string& xml();

      FieldBuffer field(size_t index) throw (exception::invalid_index);

      FieldBuffer field(std::string index) throw (exception::invalid_index);

      FieldBuffer operator[](size_t index) throw (exception::invalid_index);

      FieldBuffer operator[](std::string index) throw (exception::invalid_index);

      template <typename T>
      operator T();

      /**
      * Returns a copy of the underlying data buffer allocated
       */
      Data data();

      bool unpack(void* mem, size_t mem_octets);

      template <typename T>
      bool unpack(T& val);

      bool pack(const void* mem, size_t mem_octets);
      bool pack(const void* mem, size_t mem_octets, size_t n);

      template <typename T>
      bool pack(const T& val);
      template <typename T>
      FieldBuffer& operator=(const T& t);

      FieldBase::pointer field();
      RecordBuffer& buffer();

    protected:
      RecordBuffer& m_buffer;
      FieldBase::pointer m_field;
      std::string m_xml;

      /**
       * Unpacks the field according to the field's byte order and
       * returns the field as a 64-bit value in the host order.
       */
      bool unpack_uint64(uint64_t& ui64 );

      bool unpack_float(float& f);
      bool unpack_double(double& d);

      /**
       * Packs host ordered 64-bit value according to the field's
       * byte order specification.
       */
      bool pack_uint64(uint64_t data);
      bool pack_float(float f);
      bool pack_double(double d);

    };

  template <typename T>
  inline
  FieldBuffer::operator T()
  {
    T t;
    unpack(t);
    return t;
  }

  template <typename T>
  inline
  bool FieldBuffer::unpack(T& val)
  {
    bool b;
    float f;
    double d;
    uint64_t ui64;
    size_t length;

    // TODO finish this section for all allowed types
    switch ( m_field->type().type() )
      {
      case TYPE_INTEGER:
        b = unpack_uint64(ui64);
        val = static_cast<uint64_t>(ui64);
        return b;

      case TYPE_FLOATING:
        length = m_field->length(BITS);
        if ( ! (length == 64 || length == 32 ) )
          throw exception::type::floating_point_length();
        if (length == 64)
          {
            b = unpack_double(d);
            val = static_cast<T>(d);
          }
        else
          {
            b = unpack_float(f);
            val = static_cast<T>(f);
          }
        return b;
      case TYPE_NONE:
      case TYPE_ASCII:
      case TYPE_UTF8:
      case TYPE_BCD:
        break;
      }
    return unpack(&val, sizeof(T));
  }

  // TODO finish this section for string conversions
  // probably need to check that length%8 == 0 or throw
  template <>
  inline
  bool FieldBuffer::unpack(std::string& val)
  {
    double d;
    float f;
    uint64_t ui64;
    size_t length;
    std::ostringstream sout;

    // TODO finish this section for all allowed types
    switch ( m_field->type().type() )
      {
      case TYPE_INTEGER:
        if ( unpack_uint64(ui64) )
          {
            sout << ui64;
            val = sout.str();
            return true;
          }
        else
          return false;

      case TYPE_FLOATING:
        length = m_field->length(BITS);
        if ( ! (length == 64 || length == 32 ) )
          throw exception::type::floating_point_length();
        if (length == 64)
          {
            if ( unpack_double(d) )
              {
                sout << d;
                val = sout.str();
                return true;
              }
            else
              return false;
          }
        else
          {
            if ( unpack_float(f) )
              {
                sout << f;
                val = sout.str();
                return true;
              }
            else
              return false;
          }
        default:
          return false;
      }
    return false;
  }

  template <typename T>
  inline
  bool FieldBuffer::pack(const T& val)
  {
    uint64_t ui64;
    double d;
    float f;
    size_t length;

    switch ( m_field->type().type() )
      {
      case TYPE_INTEGER:
        ui64 = val;
        return pack_uint64(ui64);

      case TYPE_FLOATING:
        length = m_field->length(BITS);
        if ( ! (length == 64 || length == 32 ) )
          throw exception::type::floating_point_length();
        if (length == 64)
          {
            d = static_cast<double>(val);
            return pack_double(d);
          }
        else
          {
            f = static_cast<float>(val);
            return pack_float(f);
          }
      case TYPE_NONE:
      case TYPE_ASCII:
      case TYPE_UTF8:
      case TYPE_BCD:
        break;
      }

    return pack(&val, sizeof(T));
  }

  template <>
  inline
  bool FieldBuffer::pack(const std::string& val)
  {
    uint64_t ui64;
    double d;
    float f;
    size_t length;
    std::istringstream sin(val);

    switch ( m_field->type().type() )
      {
      case TYPE_INTEGER:
        sin >> ui64;
        return pack_uint64(ui64);

      case TYPE_FLOATING:
        length = m_field->length(BITS);
        if ( ! (length == 64 || length == 32 ) )
          throw exception::type::floating_point_length();
        if (length == 64)
          {
            sin >> d;
            return pack_double(d);
          }
        else
          {
            sin >> f;
            return pack_float(f);
          }
        default:
          return false;
      }

    return false;
  }

  template <typename T>
  inline
  FieldBuffer& FieldBuffer::operator=(const T& t)
  {
    pack(t);
    return *this;
  }


}

#endif
