/***************************************************************************
*   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
*   rvinyard@cs.nmsu.edu                                                  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU Lesser General Public License as        *
*   published by the Free Software Foundation version 2.1.                *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU Lesser General Public      *
*   License along with this library; if not, write to the                 *
*   Free Software Foundation, Inc.,                                       *
*   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
***************************************************************************/
#ifndef BITFLOAT_H
#define BITFLOAT_H

#include <typeinfo>
#include <stdexcept>

#include <bit/enums.h>
#include <bit/utility.h>
#include <bit/pointer.h>
#include <bit/data.h>

namespace bit
{

  template <typename F> class Float;

  class FloatingPoint
  {
    public:

      typedef BitPointer<FloatingPoint> pointer;

      FloatingPoint( ByteOrder bo=BYTEORDER_HOST ): byte_order(bo) { }

      virtual ~FloatingPoint() { }

      virtual Data as_data() const = 0;

      virtual size_t size() const = 0;

      virtual const void* voidptr() const = 0;

      virtual pointer clone() const = 0;

      /**
       * Sets the value from a data object by directly copying; byte ordering
       * of the data is not considered.
       * @param d The data object. The first n bytes are used to set the value.
       */
      virtual void set_value( const Data d ) = 0;

      virtual void set_value( const void* mem, size_t size ) = 0;

      /**
       * Sets the value from a data object
       * @param d The data object. The first n bytes are used to set the value.
       * @param databo The byte order of the data object
       */
      virtual void set_value( const Data d, ByteOrder databo ) = 0;

      template <typename T> operator T();

      template <typename T> FloatingPoint& operator = ( T f );

      virtual const std::type_info& int_type() = 0;

      ByteOrder byte_order;
  };

  template <typename F>
  class Float: public FloatingPoint {
    public:

      typedef BitPointer< Float<F> > pointer;

      Float(F v=0, ByteOrder bo=BYTEORDER_HOST): FloatingPoint(bo), value(v) { }

      Float(ByteOrder bo): FloatingPoint(bo), value(0) { }

      /**
       * Constructor that sets the value from a data object
       * @param d The data object. The first n bytes are used to set the value.
       * @param databo The byte order of the data object
       * @param bo The byte order of this integer.
       */
      Float(const Data d, ByteOrder databo=BYTEORDER_HOST, ByteOrder bo=BYTEORDER_HOST): FloatingPoint(bo) {
        this->set_value(d, databo);
      }

      static pointer create(F v=0, ByteOrder bo=BYTEORDER_HOST) { return pointer(new Float<F>(v,bo)); }

      static pointer create(ByteOrder bo) { return pointer(new Float<F>(bo)); }

      static pointer create(const Data d, ByteOrder databo=BYTEORDER_HOST, ByteOrder bo=BYTEORDER_HOST) {
        return pointer(new Float<F>(d,databo,bo));
      }

      ~Float() { }

      virtual FloatingPoint::pointer clone() const {
        pointer p = this->create(value, byte_order);
        return p;
      }

      operator F() const { return this->host(); }

      bool operator <  ( F f ) const { return this->host() <  f; }
      bool operator <= ( F f ) const { return this->host() <= f; }
      bool operator == ( F f ) const { return this->host() == f; }
      bool operator != ( F f ) const { return this->host() != f; }
      bool operator >= ( F f ) const { return this->host() >= f; }
      bool operator >  ( F f ) const { return this->host() >  f; }

      Float<F>& operator = ( F f )
      {
        value = f;
        this->convert_value_from_host_to_type();
        return *this;
      }

      const Float<F>& operator + ( F f ) const
      {
        this->convert_value_from_type_to_host();
        value += f;
        this->convert_value_from_host_to_type();
        return *this;
      }

      const Float<F>& operator - ( F f ) const
      {
        this->convert_value_from_type_to_host();
        value -= f;
        this->convert_value_from_host_to_type();
        return *this;
      }

      const Float<F>& operator * ( F f ) const
      {
        this->convert_value_from_type_to_host();
        value *= f;
        this->convert_value_from_host_to_type();
        return *this;
      }

      const Float<F>& operator / ( F f ) const
      {
        this->convert_value_from_type_to_host();
        value /= f;
        this->convert_value_from_host_to_type();
        return *this;
      }

      Float<F>& operator += ( F f )
      {
        this->convert_value_from_type_to_host();
        value += f;
        this->convert_value_from_host_to_type();
        return *this;
      }

      Float<F>& operator -= ( F f )
      {
        this->convert_value_from_type_to_host();
        value -= f;
        this->convert_value_from_host_to_type();
        return *this;
      }

      Float<F>& operator *= ( F f )
      {
        this->convert_value_from_type_to_host();
        value *= f;
        this->convert_value_from_host_to_type();
        return *this;
      }

      Float<F>& operator /= ( F f )
      {
        this->convert_value_from_type_to_host();
        value /= f;
        this->convert_value_from_host_to_type();
        return *this;
      }

      /**
       * Sets the value from a data object by directly copying; byte ordering
       * of the data is not considered.
       * @param d The data object. The first n bytes are used to set the value.
       */
      virtual void set_value( const Data d ) {
        if ( d.size() < sizeof(F) )
          throw;

        memcpy( &value, d.data(), sizeof(F) );
      }

      /**
       * Sets the value from a data object by directly copying; byte ordering
       * of the data is not considered.
       * @param d The data object. The first n bytes are used to set the value.
       */
      virtual void set_value( const void* mem, size_t size ) {
        if ( size < sizeof(F) )
          throw;

        memcpy( &value, mem, sizeof(F) );
      }

      /**
       * Sets the value from a data object
       * @param d The data object. The first n bytes are used to set the value.
       * @param databo The byte order of the data object
       */
      virtual void set_value( const Data d, ByteOrder databo ) {
        if ( d.size() < sizeof(F) )
          throw;

        memcpy( &value, d.data(), sizeof(F) );

        switch ( byte_order ) {
          case BYTEORDER_HOST:
            switch ( databo ) {
              case BYTEORDER_HOST:
                break;
              case BYTEORDER_NETWORK:
              case BYTEORDER_BIG_ENDIAN:
                value = be_to_host(value);
                break;
              case BYTEORDER_LITTLE_ENDIAN:
                value = le_to_host(value);
                break;
            }
            break;
          case BYTEORDER_NETWORK:
          case BYTEORDER_BIG_ENDIAN:
            switch ( databo ) {
              case BYTEORDER_HOST:
                value = host_to_net(value);
                break;
              case BYTEORDER_NETWORK:
              case BYTEORDER_BIG_ENDIAN:
                break;
              case BYTEORDER_LITTLE_ENDIAN:
                value = le_to_net(value);
                break;
            }
            break;
          case BYTEORDER_LITTLE_ENDIAN:
            switch ( databo ) {
              case BYTEORDER_HOST:
                value = host_to_le(value);
                break;
              case BYTEORDER_NETWORK:
              case BYTEORDER_BIG_ENDIAN:
                value = net_to_le(value);
                break;
              case BYTEORDER_LITTLE_ENDIAN:
                break;
            }
            break;
        }
      }

      virtual Data as_data() const
      {
        Data d( sizeof(F) );
        memcpy( d.data(), &value, sizeof(F) );
        return d;
      }

      virtual size_t size() const { return sizeof(F); }

      virtual const void* voidptr() const { return &value; }

      virtual const std::type_info& int_type() { return typeid(F); }

      F host() const {
        switch ( byte_order )
        {
          case BYTEORDER_HOST:
            return value;
          case BYTEORDER_NETWORK:
          case BYTEORDER_BIG_ENDIAN:
            return net_to_host(value);
          case BYTEORDER_LITTLE_ENDIAN:
            return le_to_host(value);
        }
        throw;
      }

      F network() const {
        switch ( byte_order )
        {
          case BYTEORDER_HOST:
            return host_to_net(value);
          case BYTEORDER_NETWORK:
          case BYTEORDER_BIG_ENDIAN:
            return value;
          case BYTEORDER_LITTLE_ENDIAN:
            return le_to_net(value);
        }
        throw;
      }

      F big_endian() const {
        return this->network();
      }

      F little_endian() const {
        switch ( byte_order )
        {
          case BYTEORDER_HOST:
            return host_to_le(value);
          case BYTEORDER_NETWORK:
          case BYTEORDER_BIG_ENDIAN:
            return net_to_le(value);
          case BYTEORDER_LITTLE_ENDIAN:
            return value;
        }
        throw;
      }

      F value;

    protected:
      void convert_value_from_type_to_host()
      {
        switch ( byte_order )
        {
          case BYTEORDER_HOST:
            return;
          case BYTEORDER_NETWORK:
          case BYTEORDER_BIG_ENDIAN:
            value = net_to_host(value);
          case BYTEORDER_LITTLE_ENDIAN:
            value = le_to_host(value);
        }
      }

      void convert_value_from_host_to_type()
      {
        switch ( byte_order )
        {
          case BYTEORDER_HOST:
            return;
          case BYTEORDER_NETWORK:
          case BYTEORDER_BIG_ENDIAN:
            value = host_to_net(value);
          case BYTEORDER_LITTLE_ENDIAN:
            value = host_to_le(value);
        }
      }

  };

  typedef Float<float> FloatSingle;
  typedef Float<double> FloatDouble;
  typedef Float<long double> FloatQuadruple;

  template <typename T> FloatingPoint::operator T()
  {
    switch ( this->size() )
    {
      case 4:
      {
        FloatSingle* f;
        f = dynamic_cast<FloatSingle*>(this);
        return static_cast<T>(f->host());
      }
      case 8:
      {
        FloatDouble* f;
        f = dynamic_cast<FloatDouble*>(this);
        return static_cast<T>(f->host());
      }
      case 16:
      {
        FloatQuadruple* f;
        f = dynamic_cast<FloatQuadruple*>(this);
        return static_cast<T>(f->host());
      }
    }
    throw std::logic_error("bit::FloatingPoint bad cast");
  }

  template <typename T> FloatingPoint& FloatingPoint::operator = ( T i )
  {
    switch ( this->size() )
    {
      case 4:
      {
        FloatSingle* f;
        f = dynamic_cast<FloatSingle*>(this);
        *f = (float)(i);
      }
        break;
      case 8:
      {
        FloatDouble* f;
        f = dynamic_cast<FloatDouble*>(this);
        *f = (double)(i);
      }
      break;
      case 16:
      {
        FloatQuadruple* f;
        f = dynamic_cast<FloatQuadruple*>(this);
        *f = (long double)(i);
      }
      break;
    }
    return *this;
  }

}

#endif
